# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Runs a mock server."""

from typing import Sequence

from absl import app
from absl import flags

from jetstream.core.implementations.mock import config as mock_config
from jetstream.core import server_lib


_PORT = flags.DEFINE_integer("port", 9000, "port to listen on")
_CONFIG = flags.DEFINE_string(
    "config",
    "InterleavedCPUTestServer",
    "available servers",
)


def main(argv: Sequence[str]):
  del argv
  # No devices for local cpu test. A None for prefill and a None for generate.
  devices = server_lib.get_devices()
  server_config = mock_config.get_server_config(_CONFIG.value)
  # We separate credential from run so that we can unit test it with local
  # credentials.
  # TODO: Add grpc credentials for OSS.
  jetstream_server = server_lib.run(
      port=_PORT.value,
      config=server_config,
      devices=devices,
  )
  jetstream_server.wait_for_termination()


if __name__ == "__main__":
  app.run(main)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jetstream/engine/tokenizer.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n jetstream/engine/tokenizer.proto\x12\x06\x65ngine"6\n\x13TokenizerParameters\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\textra_ids\x18\x02 \x01(\x05\x42\x02P\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "jetstream.engine.tokenizer_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals["DESCRIPTOR"]._options = None
  _globals["DESCRIPTOR"]._serialized_options = b"P\001"
  _globals["_TOKENIZERPARAMETERS"]._serialized_start = 44
  _globals["_TOKENIZERPARAMETERS"]._serialized_end = 98
# @@protoc_insertion_point(module_scope)



"""ReturnSample is a data structure utility.

It is a data structure that stores the return samples.
"""

import dataclasses


@dataclasses.dataclass
class ReturnSample:
  """Both the token ids, their string representation, and other data.

  Attributes:
    text: Text piece(s) detokenized from token id(s).
    token_ids: Raw result token id(s).
  """

  text: list[str]
  token_ids: list[int]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Proxy util functions."""

import dataclasses
import logging
import jax
import time
from fastapi import FastAPI
import uvicorn


# TODO: add a manner way to terminate.
def start_profiling_server(port: int):

  logging.info("Starting JAX profiler server on port %s", port)
  app = FastAPI()

  @dataclasses.dataclass
  class ProfilingConfig:
    seconds: int
    output_dir: str

  @app.post("/profiling")
  async def profiling(pc: ProfilingConfig):
    jax.profiler.start_trace(pc.output_dir)
    logging.info("Capturing the profiling data for next %s seconds", pc.seconds)
    time.sleep(pc.seconds)
    logging.info("Writing profiling data to %s", pc.output_dir)
    jax.profiler.stop_trace()
    return {"response": "profiling completed"}

  @app.get("/")
  async def root():
    return {"message": "Hello from proxy profiling server"}

  uvicorn.run(app, host="0.0.0.0", port=port, log_level="info")



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import transformers
from keras_hub.src.utils.preset_utils import load_json
from kithara.model import supported_models


gemma2_2b_config = transformers.Gemma2Config(
    num_hidden_layers=26,
    num_attention_heads=8,
    num_key_value_heads=4,
    hidden_size=2304,
    intermediate_size=9216,
)

gemma2_9b_config = transformers.Gemma2Config(
    num_hidden_layers=42,
    num_attention_heads=16,
    num_key_value_heads=8,
    hidden_size=3584,
    intermediate_size=14336,
    final_logit_softcapping=30.0,
    attn_logit_softcapping=50.0,
    head_dim=256,
    sliding_window=4096,
    query_pre_attn_scalar=224,
)

gemma2_27b_config = transformers.Gemma2Config(
    num_hidden_layers=46,
    num_attention_heads=32,
    num_key_value_heads=16,
    hidden_size=4608,
    intermediate_size=36864,
    final_logit_softcapping=30.0,
    attn_logit_softcapping=50.0,
    head_dim=128,
    sliding_window=4096,
    query_pre_attn_scalar=144,
)

llama31_8b_config = transformers.LlamaConfig(
    vocab_size=128256,
    hidden_size=4096,
    intermediate_size=14336,
    num_hidden_layers=32,
    num_attention_heads=32,
    num_key_value_heads=8,
    max_position_embeddings=131072,
    rms_norm_eps=1e-5,
    bos_token_id=128000,
    eos_token_id=128001,
    attention_bias=False,
    attention_dropout=0.0,
    hidden_act="silu",
    initializer_range=0.02,
    mlp_bias=False,
    model_type="llama",
    pretraining_tp=1,
    rope_scaling={
        "factor": 8.0,
        "low_freq_factor": 1.0,
        "high_freq_factor": 4.0,
        "original_max_position_embeddings": 8192,
        "rope_type": "llama3"
    },
    rope_theta=500000.0,
    tie_word_embeddings=False,
    use_cache=True,
)

llama31_70b_config = transformers.LlamaConfig(
    vocab_size=128256,
    hidden_size=8192,
    intermediate_size=28672,
    num_hidden_layers=80,
    num_attention_heads=64,
    head_dim=128,
    num_key_value_heads=8,
    max_position_embeddings=131072,
    rms_norm_eps=1e-05,
    bos_token_id=128000,
    eos_token_id=[128001, 128008, 128009],
    rope_scaling={
        "factor": 8.0,
        "high_freq_factor": 4.0,
        "low_freq_factor": 1.0,
        "original_max_position_embeddings": 8192,
        "rope_type": "llama3",
    },
    rope_theta=500000.0,
    tie_word_embeddings=False,
)

llama31_405b_config = transformers.LlamaConfig(
    vocab_size=128256,
    hidden_size=16384,
    intermediate_size=53248,
    num_hidden_layers=126,
    num_attention_heads=128,
    num_key_value_heads=8,
    head_dim=128,
    max_position_embeddings=131072,
    rms_norm_eps=1e-05,
    bos_token_id=128000,
    eos_token_id=128001,
)

llama32_1b_config = transformers.LlamaConfig(
    vocab_size=128256,
    hidden_size=2048,
    intermediate_size=8192,
    num_hidden_layers=16,
    num_attention_heads=32,
    num_key_value_heads=8,
    max_position_embeddings=131072,
    rms_norm_eps=1e-5,
    bos_token_id=128000,
    eos_token_id=128001,
    attention_bias=False,
    attention_dropout=0.0,
    hidden_act="silu",
    initializer_range=0.02,
    mlp_bias=False,
    model_type="llama",
    pretraining_tp=1,
    rope_scaling={
        "factor": 32.0,
        "low_freq_factor": 1.0,
        "high_freq_factor": 4.0,
        "original_max_position_embeddings": 8192,
        "rope_type": "llama3",
    },
    rope_theta=500000.0,
    tie_word_embeddings=True,
)

llama32_3b_config = transformers.LlamaConfig(
    vocab_size=128256,
    hidden_size=3072,
    intermediate_size=8192,
    num_hidden_layers=28,
    num_attention_heads=24,
    num_key_value_heads=8,
    max_position_embeddings=131072,
    rms_norm_eps=1e-5,
    bos_token_id=128000,
    eos_token_id=128001,
    attention_bias=False,
    attention_dropout=0.0,
    hidden_act="silu",
    initializer_range=0.02,
    mlp_bias=False,
    model_type="llama",
    pretraining_tp=1,
    rope_scaling={
        "factor": 32.0,
        "low_freq_factor": 1.0,
        "high_freq_factor": 4.0,
        "original_max_position_embeddings": 8192,
        "rope_type": "llama3",
    },
    rope_theta=500000.0,
    tie_word_embeddings=True,
    use_cache=True,
)

def get_model_name_from_preset_handle(preset_handle):
    # TODO(wenxindongwork): Support parsing presets other than HF

    config = load_json(preset_handle)
    model_type = config["model_type"]
    if model_type == "gemma2":
        n_layers = config["num_hidden_layers"]
        if n_layers == 26:
            return supported_models.GEMMA2_2B
        elif n_layers == 42:
            return supported_models.GEMMA2_9B
        elif n_layers == 46:
            return supported_models.GEMMA2_27B
    elif model_type == "llama":
        n_layers = config["num_hidden_layers"]
        if n_layers == 32:
            return supported_models.LLAMA31_8B
        elif n_layers == 80:
            return supported_models.LLAMA31_70B
        elif n_layers == 126:
            return supported_models.LLAMA31_405B
        elif n_layers == 16:
            return supported_models.LLAMA32_1B
        elif n_layers == 28:
            return supported_models.LLAMA32_3B
    print(f"model type {model_type} is currently unsupported.")
    return None


MODEL_CONFIGS = {
    supported_models.GEMMA2_2B: gemma2_2b_config,
    supported_models.GEMMA2_9B: gemma2_9b_config,
    supported_models.GEMMA2_27B: gemma2_27b_config,
    supported_models.LLAMA31_8B: llama31_8b_config,
    supported_models.LLAMA31_70B: llama31_70b_config,
    supported_models.LLAMA31_405B: llama31_405b_config,
    supported_models.LLAMA32_1B: llama32_1b_config,
    supported_models.LLAMA32_3B: llama32_3b_config,
}



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jetstream/core/proto/jetstream.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n$jetstream/core/proto/jetstream.proto\x12\x0fjetstream_proto"\x91\x03\n\rDecodeRequest\x12\x12\n\nmax_tokens\x18\x04 \x01(\x05\x12\x42\n\x0ctext_content\x18\x05 \x01(\x0b\x32*.jetstream_proto.DecodeRequest.TextContentH\x00\x12\x44\n\rtoken_content\x18\x06 \x01(\x0b\x32+.jetstream_proto.DecodeRequest.TokenContentH\x00\x12;\n\x08metadata\x18\x07 \x01(\x0b\x32\'.jetstream_proto.DecodeRequest.MetadataH\x01\x12\x13\n\x0bnum_samples\x18\x08 \x01(\x05\x1a\x1b\n\x0bTextContent\x12\x0c\n\x04text\x18\x01 \x01(\t\x1a!\n\x0cTokenContent\x12\x11\n\ttoken_ids\x18\x01 \x03(\x05\x1a\x1e\n\x08Metadata\x12\x12\n\nstart_time\x18\x01 \x01(\x02\x42\t\n\x07\x63ontentB\x13\n\x11metadata_optionalJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04"\xcb\x02\n\x0e\x44\x65\x63odeResponse\x12I\n\x0finitial_content\x18\x02 \x01(\x0b\x32..jetstream_proto.DecodeResponse.InitialContentH\x00\x12G\n\x0estream_content\x18\x03 \x01(\x0b\x32-.jetstream_proto.DecodeResponse.StreamContentH\x00\x1a\x10\n\x0eInitialContent\x1a\x81\x01\n\rStreamContent\x12\x45\n\x07samples\x18\x01 \x03(\x0b\x32\x34.jetstream_proto.DecodeResponse.StreamContent.Sample\x1a)\n\x06Sample\x12\x0c\n\x04text\x18\x01 \x01(\t\x12\x11\n\ttoken_ids\x18\x02 \x03(\x05\x42\t\n\x07\x63ontentJ\x04\x08\x01\x10\x02"\x14\n\x12HealthCheckRequest"&\n\x13HealthCheckResponse\x12\x0f\n\x07is_live\x18\x01 \x01(\x08\x32\xb9\x01\n\x0cOrchestrator\x12M\n\x06\x44\x65\x63ode\x12\x1e.jetstream_proto.DecodeRequest\x1a\x1f.jetstream_proto.DecodeResponse"\x00\x30\x01\x12Z\n\x0bHealthCheck\x12#.jetstream_proto.HealthCheckRequest\x1a$.jetstream_proto.HealthCheckResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "jetstream.core.proto.jetstream_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals["_DECODEREQUEST"]._serialized_start = 58
  _globals["_DECODEREQUEST"]._serialized_end = 459
  _globals["_DECODEREQUEST_TEXTCONTENT"]._serialized_start = 315
  _globals["_DECODEREQUEST_TEXTCONTENT"]._serialized_end = 342
  _globals["_DECODEREQUEST_TOKENCONTENT"]._serialized_start = 344
  _globals["_DECODEREQUEST_TOKENCONTENT"]._serialized_end = 377
  _globals["_DECODEREQUEST_METADATA"]._serialized_start = 379
  _globals["_DECODEREQUEST_METADATA"]._serialized_end = 409
  _globals["_DECODERESPONSE"]._serialized_start = 462
  _globals["_DECODERESPONSE"]._serialized_end = 793
  _globals["_DECODERESPONSE_INITIALCONTENT"]._serialized_start = 628
  _globals["_DECODERESPONSE_INITIALCONTENT"]._serialized_end = 644
  _globals["_DECODERESPONSE_STREAMCONTENT"]._serialized_start = 647
  _globals["_DECODERESPONSE_STREAMCONTENT"]._serialized_end = 776
  _globals["_DECODERESPONSE_STREAMCONTENT_SAMPLE"]._serialized_start = 735
  _globals["_DECODERESPONSE_STREAMCONTENT_SAMPLE"]._serialized_end = 776
  _globals["_HEALTHCHECKREQUEST"]._serialized_start = 795
  _globals["_HEALTHCHECKREQUEST"]._serialized_end = 815
  _globals["_HEALTHCHECKRESPONSE"]._serialized_start = 817
  _globals["_HEALTHCHECKRESPONSE"]._serialized_end = 855
  _globals["_ORCHESTRATOR"]._serialized_start = 858
  _globals["_ORCHESTRATOR"]._serialized_end = 1043
# @@protoc_insertion_point(module_scope)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Simple NN Module.

TODO: migrate to Flax NNX.
"""

import logging
from typing import Any
import jax
from jax import numpy as jnp
from jax.sharding import NamedSharding
from inference.nn.parameter import Parameter


class Module:
  """Simple NN module."""

  def __init__(self):
    self._parameters: dict[str, Parameter] = {}
    self._submodules: dict[str, Module] = {}

  def __setattr__(self, name: str, value: Any):
    if isinstance(value, Parameter):
      self._parameters[name] = value
    elif isinstance(value, Module | ModuleList):
      self._submodules[name] = value
    else:
      self.__dict__[name] = value

  def __getattr__(self, name: str):
    if name in self._parameters:
      return self._parameters[name]
    elif name in self._submodules:
      return self._submodules[name]
    elif name in self.__dict__:
      return self.__dict__[name]
    return None

  def init_weights(self):
    res = {}
    rng = jax.random.key(0)
    for k, param in self._parameters.items():
      param.value = jax.random.uniform(rng, param.shape, dtype=jnp.bfloat16)
      param.to_device()
      res[k] = param.value

    for k, module in self._submodules.items():
      res[k] = module.init_weights()
    return res

  def load_weights_dict(self, weights_dict):
    res = {}
    for k, v in weights_dict.items():
      attr = getattr(self, k)
      if isinstance(attr, Parameter):
        param = self._parameters[k]
        if v.shape != param.shape:
          logging.warning(
              f"Not matched shape"
              + f": defined {param.shape},"
              + f"loaded {v.shape} for module {self.__class__}"
          )
        param.value = v
        if isinstance(param.sharding, NamedSharding):
          param.to_device()
        res[k] = param.value
      elif isinstance(attr, Module) or isinstance(attr, ModuleList):
        sub_weights_dict = self._submodules[k].load_weights_dict(v)
        res[k] = sub_weights_dict
      else:
        logging.warning(
            f"Unknown checkpoint key {k} for module {self.__class__}"
        )
    return res

  def _weights_assignment_in_jit(self, weights_dict):
    for key in self._parameters:
      param = self._parameters[key]
      if isinstance(param, Parameter):
        param.value = weights_dict[key]
    for key, module in self._submodules.items():
      if key in weights_dict:
        module._weights_assignment_in_jit(weights_dict[key])

  def jittable_call(self, weights_dict, *args):
    self._weights_assignment_in_jit(weights_dict)
    return self(*args)

  # Following methods "__repr__", "_repr_with_indent" and "_spec" are
  # for debugging purpose which provide a clean string representation
  # for the model.
  def _spec(self) -> str:
    return ""

  def _repr_with_indent(self, indent) -> str:
    indent = indent + "  "
    if len(self._parameters) == 0 and len(self._submodules) == 0:
      return "{}"
    res = "{"
    for k, v in self._parameters.items():
      res += "\n" + (indent) + f"'{k}': {v}"

    for k, v in self._submodules.items():
      res += (
          "\n"
          + (indent)
          + f"'{k}': <{v.__class__.__name__}{v._spec()}> {v._repr_with_indent(indent)}"
      )

    res += "\n" + indent[:-2] + "}"
    return res

  def __repr__(self) -> str:
    return "\n" + self._repr_with_indent("")


class ModuleList:

  def __init__(self, modules: list[Module]) -> None:
    self._modules: dict[int, Module] = {}
    for i, m in enumerate(modules):
      self._modules[i] = m

  def __getitem__(self, key):
    return self._modules[key]

  def __setitem__(self, key, value):
    self._modules[key] = value

  def _spec(self) -> str:
    return ""

  def _repr_with_indent(self, indent) -> str:
    indent = indent + "  "
    if len(self._modules) == 0:
      return "{}"
    res = "{"

    for k, v in self._modules.items():
      res += (
          "\n"
          + (indent)
          + f"'{k}': <{v.__class__.__name__}{v._spec()}> {v._repr_with_indent(indent)}"
      )

    res += "\n" + indent[:-2] + "}"
    return res

  def __repr__(self) -> str:
    return "\n" + self._repr_with_indent("")

  def init_weights(self):
    res = {}
    for k, module in self._modules.items():
      if isinstance(k, int):
        res[k] = module.init_weights()
      else:
        logging.warning(f"Unknown checkpoint key {k} for module list")
    return res

  def load_weights_dict(self, weights_dict):
    res = {}
    for k, v in weights_dict.items():
      if isinstance(k, int):
        ws = self._modules[k].load_weights_dict(v)
        res[k] = ws
      else:
        logging.warning(f"Unknown checkpoint key {k} for module list")
    return res

  def _weights_assignment_in_jit(self, weights_dict):
    for layer_num, module in self._modules.items():
      module._weights_assignment_in_jit(weights_dict[layer_num])


class Model(Module):

  def jittable_call(
      self,
      weights_dict,
      input_ids,
      positions,
      kv_caches,
      attn_metadata,
  ) -> tuple[jax.Array, list[Any]]:
    self._weights_assignment_in_jit(weights_dict)
    return self(input_ids, positions, kv_caches, attn_metadata)


class CausalLM(Module):

  def jittable_call(
      self,
      weights_dict,
      input_ids: jax.Array,
      positions: jax.Array,
      kv_caches: Any,
      attn_metadata: Any,
      sampling_params: Any,
  ) -> tuple[jax.Array, list[Any]]:
    self._weights_assignment_in_jit(weights_dict)
    return self(
        input_ids,
        positions,
        kv_caches,
        attn_metadata,
        sampling_params,
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miniature load test of the mock server."""

import concurrent.futures
import functools
import time
from typing import Iterator, Sequence

from absl import app
from absl import flags
import grpc
from jetstream.core.proto import jetstream_pb2
from jetstream.core.proto import jetstream_pb2_grpc


_SERVER = flags.DEFINE_string("server", "0.0.0.0", "server address")
_PORT = flags.DEFINE_string("port", "9000", "port to ping")
_TEXT = flags.DEFINE_string("text", "AB", "The message")
_MAX_TOKENS = flags.DEFINE_integer(
    "max_tokens", 100, "Maximum number of output/decode tokens of a sequence"
)


def collect_tokens(
    response: Iterator[jetstream_pb2.DecodeResponse], print_interim: bool
) -> list[str]:
  tokens = []
  for resp in response:
    text_pieces = resp.stream_content.samples[0].text
    if print_interim:
      print(text_pieces, end="", flush=True)
    tokens.extend(text_pieces)
  return tokens


def api_call(
    stub: jetstream_pb2_grpc.OrchestratorStub,
    text: str,
    max_tokens: int,
    print_interim: bool = True,
) -> str:
  """Sends a request to server and returns text."""
  request = jetstream_pb2.DecodeRequest(
      text_content=jetstream_pb2.DecodeRequest.TextContent(text=text),
      max_tokens=max_tokens,
  )
  response = stub.Decode(request)
  print("---------------------- Sent!!!----------------------")
  tokens = collect_tokens(response, print_interim=print_interim)

  return "".join(tokens)


def ping(
    stub: jetstream_pb2_grpc.OrchestratorStub, text: str, number: int
) -> str:
  response = api_call(stub, text, _MAX_TOKENS.value, print_interim=False)
  print(f"Completed {number}")
  return response


def load_test(
    stub: jetstream_pb2_grpc.OrchestratorStub,
    text: list[str],
    queries: int = 64,
) -> list[str]:
  """Sends many queries to the server."""
  assert queries % len(text) == 0
  # repeat out
  text = text * (queries // len(text))
  number = list(range(len(text)))
  start = time.time()
  ping_partial = functools.partial(ping, stub)
  with concurrent.futures.ThreadPoolExecutor(max_workers=queries) as executor:
    responses = list(executor.map(ping_partial, text, number))
  time_taken = time.time() - start
  print(f"Time taken: {time_taken}")
  print(f"QPS: {queries/time_taken}")
  return responses


def main(argv: Sequence[str]):
  del argv
  address = f"{_SERVER.value}:{_PORT.value}"
  # Note: Uses insecure_channel only for local testing. Please add grpc
  # credentials for Production.
  with grpc.insecure_channel(address) as channel:
    grpc.channel_ready_future(channel).result()
    stub = jetstream_pb2_grpc.OrchestratorStub(channel)
    _ = load_test(stub, text=[_TEXT.value], queries=64)


if __name__ == "__main__":
  app.run(main)



"""
 Copyright 2025 Google LLC

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

      https://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """

import datetime
import sys
import os
import args_helper as helper

parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(parent_dir)

import maxtext_trillium_model_configs as model_configs
import maxtext_xpk_runner as mxr

PROXY_IMAGE = "us-docker.pkg.dev/cloud-tpu-v2-images-dev/pathways/unsanitized_proxy_server:latest"
SERVER_IMAGE = "us-docker.pkg.dev/cloud-tpu-v2-images-dev/pathways/unsanitized_server:latest"
RUNNER = "us-docker.pkg.dev/cloud-tpu-v2-images-dev/pathways/maxtext_jax_stable:latest"

# Cluster Params
CLUSTER = "v6e-256-cluster"
PROJECT = "tpu-prod-env-cluster"
ZONE = "us-east5-b"
REGION = "us-east5"
COUNTRY = "us"
DEVICE_TYPE = "v6e-256"

# Other parameters (MUST BE SET BY USER)
XPK_PATH = "../xpk"  # We're running this script from the maxtext directory
USER = os.environ["USER"]
BASE_OUTPUT_DIRECTORY = (
    f"gs://{USER}-{PROJECT}-{COUNTRY}/pw_mcjax_benchmarking/"
)

MAX_RESTARTS = 10_000
BENCHMARK_STEPS=10_000_000


def main() -> int:
  # V6e cluster config
  cluster_config = mxr.XpkClusterConfig(
      cluster_name=CLUSTER,
      project=PROJECT,
      zone=ZONE,
      device_type=DEVICE_TYPE,
  )

  # Handle command line arguments using args_helper
  should_continue = helper.handle_cmd_args(
      cluster_config, helper.DELETE, xpk_path=XPK_PATH
  )

  if not should_continue:
    return 0

  model_list = [
      # model_configs.llama3_1_70b_8192_pw_lr_real_data,
      # model_configs.llama3_1_8b_8192,
      # model_configs.llama3_1_70b_8192_iter_synth_data_and_checkpointing,
      model_configs.llama3_1_70b_8192_iter_real_data_and_checkpointing_tfds,
  ]
  pathways_config = mxr.PathwaysConfig(
      server_image=SERVER_IMAGE,
      proxy_server_image=PROXY_IMAGE,
      runner_image=RUNNER,

      # User can add additional flags here.
      server_flags="--enable_metrics_collection=true",
      proxy_flags="--enable_metrics_collection=true",
      worker_flags="--enable_metrics_collection=true",

      # server_flags="--enable_metrics_collection=false",
      # proxy_flags="--enable_metrics_collection=false",
      # worker_flags="--enable_metrics_collection=false",
  )
  num_slices_list = [
      2
  ]

  xpk_workload_cmds = []
  xpk_workload_names = []

  for model in model_list:
    # Run workloads on the below clusters
    for cluster_config in [
        cluster_config,
    ]:

      # Make modifications to the model config here to add in any additional
      # flags or changes to the model config.
      model.tuning_params["use_vertex_tensorboard"] = True
      model.tuning_params["vertex_tensorboard_project"] = PROJECT
      model.tuning_params["vertex_tensorboard_region"] = REGION

      # Run workloads in the following slice configurations
      for num_slices in num_slices_list:
        wl_config = mxr.WorkloadConfig(
            model=model,
            num_slices=num_slices,
            device_type=cluster_config.device_type,
            base_output_directory=BASE_OUTPUT_DIRECTORY,
            max_restarts=MAX_RESTARTS,
            libtpu_type=None,
            libtpu_nightly_version="",
            base_docker_image=None,
            pathways_config=pathways_config,
            xpk_path=XPK_PATH,
            num_steps=BENCHMARK_STEPS,
            priority="medium",
        )
        command, name = mxr.generate_xpk_workload_cmd(
            cluster_config=cluster_config, wl_config=wl_config
        )

        print(f"Name of the workload is: {name} \n")
        xpk_workload_names.append(name)

        print(f"XPK command to be used is: {command} \n")
        xpk_workload_cmds.append(command)

  for xpk_workload_name, xpk_workload_cmd in zip(
      xpk_workload_names, xpk_workload_cmds
  ):
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    print(
        f"[{timestamp}] Running workload: {xpk_workload_name} with command:"
        f" {xpk_workload_cmd}"
    )
    return_code = mxr.run_command_with_updates(
        xpk_workload_cmd, xpk_workload_name
    )
    if return_code != 0:
      print(f"Unable to run xpk workload: {xpk_workload_name}")


if __name__ == "__main__":
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple test engine for the JetStream API described.

Contains simple functions that we can hand calculate the desired outcome of.

Prefill: Doubles the sequence by multiplying it with an integer weight.
Insert: Writes this sequence into a cache row.
Generate step: Return sum(prefill_cache) + sum(generate_cache)/weight.

I.e. if we prefill [2, 65, 66] (i.e. <BOS>, 'A', 'B') using an ACII vocab,
we should get [4, 130, 132].

If we then insert that and run three generation steps, we should see
266+0 / 2 = 266
266 + [266] /2  = 399
266 + [266, 399] /2 = 598
I.e. ['Ċ', 'Ə', 'ɖ'] when converted back with chr()
"""

import functools
from dataclasses import asdict
from typing import Any, Callable, Optional, Tuple

import jax
import jax.numpy as jnp
import uuid
from flax import struct
from jax.experimental import mesh_utils

from jetstream.engine import engine_api
from jetstream.engine import tokenizer_pb2

Params = jax.Array  # [1,].


@struct.dataclass
class Prefix:
  """Result of the prefill step.

  This structure currently matches the prefix structure returned
  by the MaxText prefill operation, but it is flexible and can vary
  depending on the engine implementation.
  """

  logits: Optional[jax.Array] = None
  cache: Optional[jax.Array] = None
  next_pos: Optional[int] = None
  num_generated_tokens: Optional[int] = None
  first_token: Optional[int] = None


@struct.dataclass
class DecodeState:
  """The inputs into a generation step."""

  prefill_cache: jax.Array
  generate_cache: jax.Array
  generate_cache_index: int
  generate_lengths: jax.Array
  generate_tokens: jax.Array


class TestEngine(engine_api.Engine):
  """The computational core of the generative model server.

  Engine defines an API that models must adhere to as they plug into the
  JetStream efficient serving infrastructure.
  """

  def __init__(
      self,
      batch_size: int,
      cache_length: int,
      weight: float,
      vocab_size: int = 1024,
      use_chunked_prefill: bool = False,
  ):
    self.prefill_cache_batch = batch_size
    self.generate_cache_batch = batch_size
    self.cache_length = cache_length
    self.weight = weight
    self.vocab_size = vocab_size
    self._mesh = jax.sharding.Mesh(
        mesh_utils.create_device_mesh((1, 1, 1), jax.devices()), ("x", "y", "z")
    )
    self._prng_key = jax.random.PRNGKey(42)
    self._use_chunked_prefill = use_chunked_prefill

  def load_params(self) -> Params:
    """Loads model weights."""
    # An integer, used to multiply inputs.
    return jnp.array([self.weight], dtype=jnp.float32)

  def load_params_dict(self) -> Params:
    """Loads model weights."""
    # An integer, used to multiply inputs.
    return {"params": jnp.array([self.weight], dtype=jnp.float32)}

  @functools.partial(
      jax.jit,
      static_argnums=(0,),
      static_argnames=("request_id",),
  )
  # pylint: disable=unused-argument
  def prefill(
      self,
      *,
      params: Params,
      existing_prefix: Optional[jax.Array] = None,
      padded_tokens: jax.Array,
      true_length: int,
      request_id: Optional[uuid.UUID] = None,
      previous_chunk=None,
      complete_padded_prompt=None,
      complete_prompt_true_length=None,
      positions=None,
  ) -> Tuple[Prefix, engine_api.ResultTokens]:
    """Computes a kv-cache for a new generate request.

    Args:
      params: Scalar multiplier.
      existing_prefix: If provided, represents a prefix that has already been
        processed by the underlying model.
      padded_tokens: Logically appended tokens to any existing prefix, this is
        what we compute prefill on.
      true_length: The real length of the tokens, pre-pad.
    Returns:
      kv_cache: For the resulting text.
    """
    if existing_prefix is not None:
      raise NotImplementedError
    assert padded_tokens.ndim == 1

    # Generate dummy prefill cache content
    if not self._use_chunked_prefill:
      prefill_cache = padded_tokens[None, :] * params
    else:
      prefill_cache = padded_tokens[None, :]

    # Create a dummy first generated token.
    first_generated_token = (prefill_cache.sum(axis=-1).astype(jnp.int32))[
        :, jnp.newaxis
    ]

    if not self._use_chunked_prefill:
      prefix = Prefix(
          logits=jax.random.normal(self._prng_key, (1, self.vocab_size)),
          cache=prefill_cache,
          next_pos=jnp.full((1, 1), true_length, dtype=jnp.int32),
          num_generated_tokens=jnp.zeros((1, 1), dtype=jnp.int32),
          first_token=first_generated_token,
      )
    else:
      prefix = {
          "logits": jax.random.normal(self._prng_key, (1, self.vocab_size)),
          "cache": prefill_cache,
          "next_pos": jnp.full((1, 1), true_length, dtype=jnp.int32),
          "generated_tokens": jnp.zeros((1, 1), dtype=jnp.int32),
          "tokens": first_generated_token,
          "first_token": first_generated_token,
      }

    speculations = first_generated_token.shape[1]
    result_tokens = engine_api.ResultTokens(
        data=jnp.concatenate(
            (
                first_generated_token,
                jnp.ones_like(first_generated_token),
                jnp.ones_like(first_generated_token),
            ),
            axis=-1,
        ),
        tokens_idx=(0, speculations),
        # Validity occupies the same amount of space, but next in line.
        valid_idx=(speculations, 2 * speculations),
        # And lengths is rank 1.
        length_idx=(2 * speculations, 2 * speculations + 1),
        samples_per_slot=self.generate_cache_batch // self.prefill_cache_batch,
    )
    return (prefix, result_tokens)

  @functools.partial(
      jax.jit, static_argnums=(0,), static_argnames=("num_samples",)
  )
  def prefill_multisampling(
      self,
      *,
      params: Params,
      existing_prefix: Optional[jax.Array] = None,
      padded_tokens: jax.Array,
      true_length: int,
      sampler: Optional[Callable[[Any], Any]] = None,  # pylint: disable=unused-argument
      rng: Optional[engine_api.PRNGKeyType] = None,
      num_samples: int = 1,
  ) -> Tuple[Prefix, engine_api.ResultTokens]:
    """Computes a kv-cache for a new generate request.

    With multi-sampling, the engine will generate multiple first tokens in the
    prefilling stage. The number of tokens is specified by num_samples.
    """
    if existing_prefix is not None:
      raise NotImplementedError
    assert padded_tokens.ndim == 1

    # Generate dummy prefill cache content
    prefill_cache = padded_tokens[None, :] * params

    # Create dummy first generated tokens.
    first_generated_tokens = []
    for _ in range(num_samples):
      first_generated_token = (prefill_cache.sum(axis=-1).astype(jnp.int32))[
          :, jnp.newaxis
      ]
      first_generated_tokens.append(first_generated_token)
    first_generated_tokens = jnp.concatenate(first_generated_tokens, axis=0)

    prefix = Prefix(
        logits=jax.random.normal(self._prng_key, (1, self.vocab_size)),
        cache=prefill_cache,
        next_pos=jnp.full((1, 1), true_length, dtype=jnp.int32),
        num_generated_tokens=jnp.zeros((num_samples, 1), dtype=jnp.int32),
        first_token=first_generated_tokens,
    )

    speculations = first_generated_token.shape[1]
    result_tokens = engine_api.ResultTokens(
        data=jnp.concatenate(
            (
                first_generated_tokens,
                jnp.ones_like(first_generated_tokens),
                jnp.ones_like(first_generated_tokens),
            ),
            axis=-1,
        ),
        tokens_idx=(0, speculations),
        # Validity occupies the same amount of space, but next in line.
        valid_idx=(speculations, 2 * speculations),
        # And lengths is rank 1.
        length_idx=(2 * speculations, 2 * speculations + 1),
        samples_per_slot=num_samples,
    )
    return (prefix, result_tokens)

  @functools.partial(jax.jit, static_argnums=(0,))
  def generate(
      self, params: Params, decode_state: DecodeState
  ) -> Tuple[DecodeState, engine_api.ResultTokens]:
    """Generates tokens for each sequence being decoded in parallel."""
    (
        prefill_cache,
        generate_cache,
        generate_cache_index,
        generate_lengths,
        previous_timestep,
    ) = (
        decode_state.prefill_cache,
        decode_state.generate_cache,
        decode_state.generate_cache_index,
        decode_state.generate_lengths,
        decode_state.generate_tokens,
    )

    # Update generate cache
    generate_cache = jax.lax.dynamic_update_slice_in_dim(
        generate_cache,
        previous_timestep.astype(jnp.float32),
        start_index=generate_cache_index,
        axis=1,
    )
    generate_cache_index = (generate_cache_index + 1) % self.cache_length

    # Sum each row of prefill cache and generate cache to produce new timestep,
    # multiply by params.
    l_iota = jax.lax.broadcasted_iota(
        jnp.int32,
        (self.generate_cache_batch, self.cache_length),
        dimension=1,
    )

    # The generate cache should be circular and right aligned.
    # TODO: Do we need a left aligned one to test spec sampling?
    # Don't need the + 1 you normally would, because we don't provide a
    # token from prefill in the dummy.
    # This iota and masking is to allow for a cicular cache.
    length_mask = (
        -(l_iota - generate_cache_index) % self.cache_length
    ) <= generate_lengths[:, None]
    length_masked_gen_cache = generate_cache * length_mask
    generated_tokens = (
        prefill_cache.sum(axis=-1)
        + (length_masked_gen_cache.sum(axis=-1) / params)
    )[:, jnp.newaxis]
    # Wait to simulate model step time.
    fake_size = 4096
    fake_work = jnp.ones((fake_size, fake_size)) @ jnp.ones(
        (fake_size, fake_size)
    )
    # Do some fake work that isn't eliminated by dead code elimination (DCE).
    generate_cache = generate_cache + fake_work.mean() - fake_work.mean()
    new_lengths = generate_lengths + 1
    speculations = generated_tokens.shape[1]
    # Concatenates the tokens, their validity and the lengths of each sequence
    # into one tensor so that copy operations are faster on Cloud TPU
    # infrastructure.
    token_data = jnp.concatenate(
        [
            generated_tokens,
            jnp.ones_like(generated_tokens),
            new_lengths[:, None],
        ],
        axis=-1,
    )
    return DecodeState(
        prefill_cache=prefill_cache,
        generate_cache=generate_cache,
        generate_cache_index=generate_cache_index,
        generate_lengths=new_lengths,
        generate_tokens=generated_tokens,
    ), engine_api.ResultTokens(
        data=token_data.astype(jnp.int32),
        # Tokens are shape [batch, speculations], so when we concatenate
        # tokens, validity and length along their index 1 dimension then they
        # occupy 0:speculations.
        tokens_idx=(0, speculations),
        # Validity occupies the same amount of space, but next in line.
        valid_idx=(speculations, 2 * speculations),
        # And lengths is rank 1.
        length_idx=(2 * speculations, 2 * speculations + 1),
        samples_per_slot=self.generate_cache_batch // self.prefill_cache_batch,
    )

  @functools.partial(
      jax.jit,
      static_argnums=(0,),
      donate_argnums=(2,),
      static_argnames=("request_id",),
  )
  def insert(
      self,
      prefix: Any,
      decode_state: DecodeState,
      slot: int,
      request_id: Optional[uuid.UUID] = None,
  ) -> DecodeState:
    """Adds `prefix` into `decode_state` at `slot`."""
    if not self._use_chunked_prefill:
      prefill_cache = prefix.cache
    else:
      prefill_cache = prefix["cache"]

    prefill_cache = jax.lax.dynamic_update_slice_in_dim(
        decode_state.prefill_cache, prefill_cache * 1.0, slot, axis=0
    )
    generate_cache = jax.lax.dynamic_update_slice_in_dim(
        decode_state.generate_cache,
        jnp.zeros((1, self.cache_length)),
        slot,
        axis=0,
    )
    samples_per_slot = self.generate_cache_batch // self.prefill_cache_batch
    generate_lengths = jax.lax.dynamic_update_slice_in_dim(
        decode_state.generate_lengths,
        jnp.ones((samples_per_slot), dtype=jnp.int32),
        slot * samples_per_slot,
        axis=0,
    )
    if not self._use_chunked_prefill:
      first_token = prefix.first_token
    else:
      first_token = prefix["first_token"]
    generate_tokens = jax.lax.dynamic_update_slice_in_dim(
        decode_state.generate_tokens,
        first_token,
        slot * samples_per_slot,
        axis=0,
    )
    return decode_state.replace(
        prefill_cache=prefill_cache,
        generate_cache=generate_cache,
        generate_lengths=generate_lengths,
        generate_tokens=generate_tokens,
    )

  @functools.partial(jax.jit, static_argnums=(0,), donate_argnums=(2,))
  def bulk_insert(
      self,
      prefix: Prefix,
      decode_state: DecodeState,
      slots: list[int],
  ) -> DecodeState:
    """Insert a single computed prefill cache into multiple slots in
    KV cache.
    """
    prefill_cache = decode_state.prefill_cache
    generate_cache = decode_state.generate_cache
    generate_lengths = decode_state.generate_lengths
    generate_tokens = decode_state.generate_tokens
    for slot in slots:
      prefill_cache = jax.lax.dynamic_update_slice_in_dim(
          prefill_cache, prefix.cache, slot, axis=0
      )
      generate_cache = jax.lax.dynamic_update_slice_in_dim(
          generate_cache,
          jnp.zeros((1, self.cache_length)),
          slot,
          axis=0,
      )
      samples_per_slot = 1
      generate_lengths = jax.lax.dynamic_update_slice_in_dim(
          generate_lengths,
          jnp.ones((samples_per_slot), dtype=jnp.int32),
          slot * samples_per_slot,
          axis=0,
      )
      generate_tokens = jax.lax.dynamic_update_slice_in_dim(
          generate_tokens,
          prefix.first_token,
          slot * samples_per_slot,
          axis=0,
      )
    return decode_state.replace(
        prefill_cache=prefill_cache,
        generate_cache=generate_cache,
        generate_lengths=generate_lengths,
        generate_tokens=generate_tokens,
    )

  def get_prefix_destination_sharding(self) -> Any:
    res = jax.tree_util.tree_map(
        lambda _: jax.sharding.NamedSharding(
            mesh=self.mesh, spec=jax.sharding.PartitionSpec()
        ),
        asdict(Prefix()),
        is_leaf=lambda _: True,
    )
    return res

  def get_tokenizer(self) -> tokenizer_pb2.TokenizerParameters:
    """Return a protobuf of tokenizer info, callable from Py or C++."""
    return tokenizer_pb2.TokenizerParameters(path="test", extra_ids=0)

  def init_decode_state(self) -> DecodeState:
    """Initialises any state which a generation step transforms."""
    return DecodeState(
        prefill_cache=jnp.zeros(
            (self.prefill_cache_batch, self.cache_length), dtype=jnp.float32
        ),
        generate_cache=jnp.zeros(
            (self.generate_cache_batch, self.cache_length), dtype=jnp.float32
        ),
        generate_cache_index=0,
        generate_lengths=jnp.zeros(
            (self.generate_cache_batch), dtype=jnp.int32
        ),
        generate_tokens=jnp.zeros(
            (self.generate_cache_batch, 1), dtype=jnp.int32
        ),
    )

  @property
  def max_concurrent_decodes(self) -> int:
    """Free slots."""
    return self.prefill_cache_batch

  @property
  def max_prefill_length(self) -> int:
    """Maximum prefill length."""
    return self.cache_length

  @property
  def samples_per_slot(self) -> int:
    """Number of samples per slot."""
    return self.generate_cache_batch // self.max_concurrent_decodes

  @property
  def mesh(self) -> jax.sharding.Mesh:
    """Mesh which the engine is running on."""
    return self._mesh

  @property
  def colocated_cpus(self) -> None:
    """CPU devices colocated with the engine's accelerators."""
    raise NotImplementedError

  @property
  def use_chunked_prefill(self) -> bool:
    """Maximum prefill length."""
    return self._use_chunked_prefill

  @property
  def prefill_chunk_size(self) -> int:
    """Maximum prefill length."""
    return 64



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests functionality of inference sampling utils."""

import jax
import jax.numpy as jnp
import unittest
from jetstream.engine import sampling_utils


class SamplingUtilsTest(unittest.TestCase):

  def setUp(self):
    self.rng = jax.random.PRNGKey(0)
    self.logits = jnp.array([[-0.5, 1.2, 0.8], [-1.0, 0.3, 0.7]])

  def test_greedy_sampling(self):
    token = sampling_utils.sampling(self.logits, self.rng, "greedy")
    expected_token = jnp.array([1, 2])
    self.assertTrue(jnp.array_equal(token, expected_token))

  def test_weighted_sampling(self):
    # Multiple samples to increase the chance of catching errors
    for _ in range(10):
      result = sampling_utils.sampling(self.logits, self.rng, "weighted")
      self.assertTrue(
          jnp.all(jnp.isin(result, jnp.array([0, 1, 2])))
      )  # Check if sampled from valid indices

  def test_nucleus_sampling(self):
    for _ in range(10):
      result = sampling_utils.sampling(
          self.logits, self.rng, "nucleus", nucleus_topp=0.8
      )
      self.assertTrue(jnp.all(jnp.isin(result, jnp.array([0, 1, 2]))))
    invalid_topp = -0.1
    with self.assertRaises(ValueError) as context:
      sampling_utils.sampling(
          self.logits, self.rng, "nucleus", nucleus_topp=invalid_topp
      )
      self.assertIn(
          f"Can't apply nucleus with parameter {invalid_topp=} less zero",
          str(context.exception),
      )

  def test_topk_sampling(self):
    for _ in range(10):
      result = sampling_utils.sampling(self.logits, self.rng, "topk", topk=2)
      self.assertTrue(
          jnp.all(jnp.isin(result, jnp.array([1, 2])))
      )  # Only top 2 logits should be sampled
    invalid_topk = 0
    with self.assertRaises(ValueError) as context:
      sampling_utils.sampling(self.logits, self.rng, "topk", topk=invalid_topk)
      self.assertIn(
          f"Can't apply algorithm topk with parameter {invalid_topk=} <= 0",
          str(context.exception),
      )

  def test_unsupported_algorithm(self):
    with self.assertRaises(ValueError):
      sampling_utils.sampling(self.logits, self.rng, "unsupported_algorithm")



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configs of engines for the orchestrator to load."""

import dataclasses
import functools
from typing import Any, Callable, List, Optional, Tuple, Type
from numpy import uint16

from jetstream.engine import engine_api
from jetstream.engine import mock_engine


Devices = Any

CreateEngineFn = Callable[[Devices], engine_api.Engine]


@dataclasses.dataclass
class ServerConfig:
  """Configs for slices to put engines on."""

  prefill_slices: Tuple[str, ...] = ()
  generate_slices: Tuple[str, ...] = ()
  interleaved_slices: Tuple[str, ...] = ()
  prefill_engine_create_fns: Tuple[CreateEngineFn, ...] = ()
  generate_engine_create_fns: Tuple[CreateEngineFn, ...] = ()
  interleaved_engine_create_fns: Tuple[CreateEngineFn, ...] = ()
  is_ray_backend: bool = False
  # Parameters for customized gc config, increase the numbers here will
  # potentially increase memory usage.
  gc_gen0_allocs: int = 60000  # default is 700, too frequent sometimes.
  gc_gen1_multipler: int = 2  # Make gen1 gc runs less frequent
  gc_gen2_multipler: int = 3  # Make gen2 gc runs less frequent


@dataclasses.dataclass
class InstantiatedEngines:
  prefill_engines: List[engine_api.Engine]
  generate_engines: List[engine_api.Engine]
  interleaved_engines: List[engine_api.Engine]


@dataclasses.dataclass
class MetricsServerConfig:
  port: uint16
  model_name: Optional[str] = None


# ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼#


def get_test_engine(devices: Devices, weight: int) -> engine_api.Engine:
  del devices
  return mock_engine.TestEngine(
      batch_size=8,
      cache_length=32,
      weight=weight,
  )


@dataclasses.dataclass
class CPUTestServer(ServerConfig):
  prefill_slices = ("cpu=1",)
  generate_slices = ("cpu=1",)
  prefill_engine_create_fns = (functools.partial(get_test_engine, weight=2),)
  generate_engine_create_fns = (functools.partial(get_test_engine, weight=4),)


@dataclasses.dataclass
class InterleavedCPUTestServer(ServerConfig):
  interleaved_slices = ("cpu=1",)
  interleaved_engine_create_fns = (
      functools.partial(get_test_engine, weight=2),
  )


# ▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼▼#


def slice_to_num_chips(s: str) -> int:
  """Converts a TPU spec like v5e-8 or v5e=8 to the number of chips, 8."""
  # Account for the case where it is written 'tpu=8' for compatibility.
  delim = "-" if "-" in s else "="
  # TODO: Support more accelerator type check.
  accelerator_type, num_devices = s.split(delim)
  return int(num_devices) if accelerator_type != "v4" else int(num_devices) // 2


def _split_devices_by_slices(
    devices: list[Devices], slices: list[int]
) -> List[List[Devices]]:
  """Converts an ordered list of devices into slices."""
  assert sum(slices) == len(devices), f"{sum(slices)} != {len(devices)}"
  cumsum = 0
  slice_split_devices = []
  for sl in slices:
    slice_split_devices.append(devices[cumsum : cumsum + sl])
    cumsum += sl
  return slice_split_devices


def get_engines(
    server_config: Type[ServerConfig], devices: List[Devices]
) -> InstantiatedEngines:
  """Processes config to get the appropriate engines.

  Args:
    server_config: ServerConfig.
    devices: Device objects.

  Returns:
    Instantiated engines!

  Devices are popped in order!
  """
  # Now, we need to split devices by slice due to TPU backend config.
  slices: list[int] = [
      slice_to_num_chips(s)
      for s in list(server_config.prefill_slices)
      + list(server_config.generate_slices)
      + list(server_config.interleaved_slices)
  ]
  if sum(slices) != len(devices):
    raise ValueError(
        f"The number of available devices ({len(devices)}) do not match the "
        f"expected number of devices on all the slices ({sum(slices)}) "
        "specified in the server_config:\n"
        f"{server_config.prefill_slices=}\n"
        f"{server_config.generate_slices=}\n"
        f"{server_config.interleaved_slices=}\n"
    )
  # e.g. [[tpu_0], [tpu_1]] corresponding to prefill: v5e=1x1
  # generate = v5e=1x1; or [[tpu_0, tpu_1, tpu_2, tpu_3]] corresponding to
  # interleaved: v5e=2x2
  split_devices = _split_devices_by_slices(devices, slices)
  prefill_engines = [
      e(split_devices.pop(0)) for e in server_config.prefill_engine_create_fns
  ]
  generate_engines = [
      e(split_devices.pop(0)) for e in server_config.generate_engine_create_fns
  ]
  # These share chips and weights for prefill and generation.
  interleaved_engines = [
      e(split_devices.pop(0))
      for e in server_config.interleaved_engine_create_fns
  ]
  return InstantiatedEngines(
      prefill_engines, generate_engines, interleaved_engines
  )



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""batch scheduler"""

import enum
import queue
import dataclasses
from inference.runtime.kv_cache import KVCacheManager
from inference.runtime.request_type import PrefillRequest, GenerateRequest, GenerateState


@dataclasses.dataclass
class PrefillPagesUpdate:
  page_indices: list[int]  # length: chunk_size // page_size


@dataclasses.dataclass
class GenerateStatePageUpdate:
  slot: int
  page_idx: int
  mapped_idx: int


@dataclasses.dataclass
class Schedule:
  schedule_prefill: bool
  prefill_request: PrefillRequest
  prefill_pages_update: PrefillPagesUpdate
  schedule_generate: bool
  new_generate_requests: list[GenerateRequest]
  generate_state_page_updates: list[GenerateStatePageUpdate]


@enum.unique
class SchedulePolicy(enum.Enum):
  OFFLINE = enum.auto()
  ONLINE = enum.auto()


class BatchScheduler:

  def __init__(
      self,
      kv_cache_manager: KVCacheManager,
      batch_size: int,
      max_seq_len: int,
      schedule_policy: SchedulePolicy = SchedulePolicy.OFFLINE,
  ):
    self.prefill_queue: queue.Queue[PrefillRequest] = queue.Queue()
    self.generate_queue: queue.Queue[GenerateRequest] = queue.Queue()
    self.kv_manager = kv_cache_manager
    self.batch_size = batch_size
    self.max_seq_len = max_seq_len
    self.schedule_policy = schedule_policy

  def enqueue_prefill_req(self, req: PrefillRequest):
    self.prefill_queue.put(req)

  def enqueue_generate_req(self, req: GenerateRequest):
    self.generate_queue.put(req)

  def schedule(
      self,
      active_prefill: PrefillRequest | None,
      generate_state: GenerateState,
  ) -> Schedule | None:
    """Schedule the workload for next iteration. Only host state is
    updated in the schedule function.
    """
    avail_slot_size = generate_state.available_slots.qsize()
    next_prefill_req = active_prefill
    prefill_pages_update = None
    next_generate_reqs = []
    generate_state_page_updates = []

    schedule_prefill = False
    schedule_generate = False

    # Schedule new prefill req, if no active prefill request.
    if not next_prefill_req:
      if avail_slot_size > 0:
        try:
          next_prefill_req = self.prefill_queue.get_nowait()
          if not next_prefill_req:
            return None
        except queue.Empty:
          pass

    if next_prefill_req:
      cur_prompt_chunk_len = next_prefill_req.chunk_size
      total_len = len(next_prefill_req.unpadded_token_ids)
      if (
          total_len
          <= (next_prefill_req.chunk_idx + 1) * next_prefill_req.chunk_size
      ):
        cur_prompt_chunk_len = (
            total_len - next_prefill_req.chunk_idx * next_prefill_req.chunk_size
        )
      alloced_pages = self.kv_manager.alloc_prefill_hbm_pages(
          cur_prompt_chunk_len
      )
      if len(alloced_pages) == 0:
        # TODO: introduce priority for the request and better
        # eviction algorithm.
        raise NotImplementedError("Eviction is not supported yet")
      else:
        start_idx = (
            next_prefill_req.chunk_idx * next_prefill_req.chunk_size
        ) // self.kv_manager.page_size
        for i, page in enumerate(alloced_pages):
          next_prefill_req.page_indices[start_idx + i] = page
        prefill_pages_update = PrefillPagesUpdate(alloced_pages)

    # Schedule new generate reqs and allocate memory for all reqs.
    with generate_state.map_mutex:
      if (
          self.schedule_policy == SchedulePolicy.ONLINE
          or not next_prefill_req
          or (
              len(generate_state.active_slot_req_map)
              + self.generate_queue.qsize()
              > 0.95 * self.batch_size
          )
      ):
        # Add new generate request to the slots.
        while (
            generate_state.available_slots.qsize() > 0
            and self.generate_queue.qsize() > 0
        ):
          gr = self.generate_queue.get_nowait()
          if not gr:
            return None
          slot = generate_state.available_slots.get_nowait()
          gr.slot = slot
          generate_state.active_slot_req_map[slot] = gr
          next_generate_reqs.append(gr)

        # Check and alloc memory for generate.
        alloced_pages = self.kv_manager.alloc_hbm_pages(
            len(generate_state.active_slot_req_map)
        )
        if (
            len(generate_state.active_slot_req_map) != 0
            and len(alloced_pages) == 0
        ):
          raise NotImplementedError(
              "Eviction isn't supported yet, please set a lower value for batch_size"
          )

        page_to_use = 0
        for slot, req in generate_state.active_slot_req_map.items():
          idx = req.pos // self.kv_manager.page_size
          if req.pos % self.kv_manager.page_size != 0:
            continue
          if idx >= len(req.page_indices):
            continue

          req.page_indices[idx] = alloced_pages[page_to_use]
          generate_state_page_updates.append(
              GenerateStatePageUpdate(
                  slot=slot,
                  page_idx=idx,
                  mapped_idx=alloced_pages[page_to_use],
              )
          )
          page_to_use += 1

        self.kv_manager.free_hbm_pages(alloced_pages[page_to_use:])

        if len(generate_state.active_slot_req_map) == 0:
          schedule_generate = False
        else:
          schedule_generate = True

    if next_prefill_req:
      schedule_prefill = True
    else:
      schedule_prefill = False

    if not schedule_prefill and not schedule_generate:
      # Nothing got scheduled, busy waiting for either prefill
      # or generate queue to have pending request.
      while True:
        if self.prefill_queue.qsize() > 0 or self.generate_queue.qsize() > 0:
          return self.schedule(active_prefill, generate_state)

    return Schedule(
        schedule_prefill=schedule_prefill,
        prefill_request=next_prefill_req,
        prefill_pages_update=prefill_pages_update,
        schedule_generate=schedule_generate,
        new_generate_requests=next_generate_reqs,
        generate_state_page_updates=generate_state_page_updates,
    )



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Llama model family."""

import jax
from transformers import LlamaConfig
from inference import nn
from inference import parallel
from .sampling.sampler import Sampler, SamplingParams
from inference.model.postprocess import *


class _LlamaFeedForward(nn.Module):

  def __init__(
      self,
      config: LlamaConfig,
      parallel_config: parallel.FeedForwardParallelConfig,
  ):
    super().__init__()
    self.config = config
    self.parallel_config = parallel_config

    hidden_size = config.hidden_size
    intermediate_size = config.intermediate_size

    gate_up_proj_parallel = parallel.LinearParallelConfig(
        mesh=parallel_config.mesh,
        parallel_type=parallel.LinearParallelType.COLUMN,
    )

    if parallel_config.enable_collective_matmul:
      gate_up_proj_parallel.collective_matmul_type = (
          parallel.CollectiveMatmulType.ALL_GATHER
      )

    self.gate_up_proj = nn.Linear(
        in_features=hidden_size,
        out_features=[intermediate_size] * 2,
        parallel_config=gate_up_proj_parallel,
    )

    down_proj_parallel = parallel.LinearParallelConfig(
        mesh=parallel_config.mesh,
        parallel_type=parallel.LinearParallelType.ROW,
    )

    if parallel_config.enable_collective_matmul:
      down_proj_parallel.collective_matmul_type = (
          parallel.CollectiveMatmulType.REDUCE_SCATTER
      )
    else:
      down_proj_parallel.reduce_output = True

    self.down_proj = nn.Linear(
        in_features=intermediate_size,
        out_features=hidden_size,
        parallel_config=down_proj_parallel,
    )

    self.silu = jax.nn.silu

  def __call__(self, x: jax.Array) -> jax.Array:
    gate, up = self.gate_up_proj(x)
    x = self.silu(gate) * up
    x = self.down_proj(x)
    return x


class _LlamaAttention(nn.Module):

  def __init__(
      self,
      config: LlamaConfig,
      parallel_config: parallel.AttentionParallelConfig,
  ):
    super().__init__()
    self.config = config
    self.parallel_config = parallel_config
    self.hidden_size = config.hidden_size
    self.num_attn_heads = config.num_attention_heads
    self.head_dim = getattr(
        config, "head_dim", self.hidden_size // self.num_attn_heads
    )
    self.num_kv_heads = config.num_key_value_heads
    self.num_kv_groups = self.num_attn_heads // self.num_kv_heads
    self.rope_theta = config.rope_theta
    self.parallel_config = parallel_config

    column_parallel_config = parallel.config.LinearParallelConfig(
        mesh=parallel_config.mesh,
        parallel_type=parallel.LinearParallelType.COLUMN,
    )

    self.q_proj = nn.Linear(
        self.hidden_size,
        self.num_attn_heads * self.head_dim,
        parallel_config=column_parallel_config,
    )

    self.k_proj = nn.Linear(
        self.hidden_size,
        self.num_kv_heads * self.head_dim,
        parallel_config=column_parallel_config,
    )

    self.v_proj = nn.Linear(
        self.hidden_size,
        self.num_kv_heads * self.head_dim,
        parallel_config=column_parallel_config,
    )

    out_proj_parallel = parallel.config.LinearParallelConfig(
        mesh=parallel_config.mesh,
        parallel_type=parallel.LinearParallelType.ROW,
    )

    if parallel_config.reduce_output:
      out_proj_parallel.reduce_output = True
    else:
      out_proj_parallel.reduce_scatter_output = True

    self.o_proj = nn.Linear(
        self.num_attn_heads * self.head_dim,
        self.hidden_size,
        out_proj_parallel,
    )

    self.rotary_emb = nn.apply_rope_embedding
    self.attn = nn.AttentionOps(
        self.num_attn_heads,
        self.num_kv_heads,
        self.head_dim,
    )

  def __call__(
      self,
      hidden_states,
      positions,
      kv_cache: nn.KVCache,
      attn_metadata: nn.AttentionMetadata,
  ) -> tuple[jax.Array, nn.KVCache]:
    if self.parallel_config.gather_input:
      hidden_states = parallel.ops.all_gather(
          hidden_states,
          len(hidden_states.shape) - 1,
          parallel.tp_major_axis_names(),
      )

    q = self.q_proj(hidden_states)
    k = self.k_proj(hidden_states)
    v = self.v_proj(hidden_states)

    # reshape as (num_tokens, num_heads, head_dim)
    q = q.reshape((q.shape[0], -1, self.head_dim))
    k = k.reshape((k.shape[0], -1, self.head_dim))
    v = v.reshape((v.shape[0], -1, self.head_dim))

    q = self.rotary_emb(q, positions, self.rope_theta)
    k = self.rotary_emb(k, positions, self.rope_theta)

    output, kv_cache = self.attn(
        q,
        k,
        v,
        kv_cache,
        attn_metadata,
    )

    output = self.o_proj(output)
    return output, kv_cache


class _LlamaDecoderLayer(nn.Module):

  def __init__(
      self,
      config: LlamaConfig,
      parallel_config: parallel.DecoderLayerParallelConfig,
  ):
    super().__init__()
    self.config = config
    self.parallel_config = parallel_config
    mesh = parallel_config.mesh

    if parallel.platform() == "tpu":
      enable_collective_matmul = True
    else:
      enable_collective_matmul = False

    self.self_attn = _LlamaAttention(
        config,
        parallel.AttentionParallelConfig(
            mesh=mesh,
            gather_input=enable_collective_matmul,
            reduce_output=(not enable_collective_matmul),
        ),
    )

    self.ffw = _LlamaFeedForward(
        config,
        parallel_config=parallel.FeedForwardParallelConfig(
            mesh, enable_collective_matmul=enable_collective_matmul
        ),
    )

    self.input_layernorm = nn.RMSNorm(
        config.hidden_size,
        eps=config.rms_norm_eps,
        parallel_config=parallel.RMSNormParallelConfig(
            mesh=mesh,
            activation_sharded=enable_collective_matmul,
        ),
    )

    self.post_attention_layernorm = nn.RMSNorm(
        config.hidden_size,
        eps=config.rms_norm_eps,
        parallel_config=parallel.RMSNormParallelConfig(
            mesh=mesh,
            activation_sharded=enable_collective_matmul,
        ),
    )

  def __call__(
      self,
      hidden_states,
      positions,
      kv_cache,
      attn_metadata: nn.AttentionMetadata,
  ) -> tuple[jax.Array, nn.KVCache]:
    residual = hidden_states

    hidden_states = self.input_layernorm(hidden_states)
    hidden_states, kv_cache = self.self_attn(
        hidden_states, positions, kv_cache, attn_metadata
    )
    hidden_states += residual

    residual = hidden_states
    hidden_states = self.post_attention_layernorm(hidden_states)

    hidden_states = self.ffw(hidden_states)
    hidden_states += residual

    return hidden_states, kv_cache


class LlamaModel(nn.Model):

  def __init__(
      self,
      config: LlamaConfig,
      parallel_config: parallel.ModelParallelConfig,
  ):
    super().__init__()
    self.config = config
    self.parallel_config = parallel_config
    self.vocab_size = config.vocab_size
    mesh = parallel_config.mesh

    self.embed_tokens = nn.Embedding(
        vocab_size=config.vocab_size,
        embedding_dim=config.hidden_size,
        parallel_config=parallel.EmbeddingParallelConfig(
            mesh=mesh,
            parallel_type=parallel.EmbeddingParallelType.COLUMN,
        ),
    )

    self.layers = nn.ModuleList(
        [
            _LlamaDecoderLayer(
                config=config,
                parallel_config=parallel.DecoderLayerParallelConfig(
                    mesh=parallel_config.mesh,
                ),
            )
            for _ in range(config.num_hidden_layers)
        ]
    )

    if parallel.platform() == "tpu":
      enable_collective_matmul = True
    else:
      enable_collective_matmul = False

    self.norm = nn.RMSNorm(
        config.hidden_size,
        config.rms_norm_eps,
        parallel_config=parallel.RMSNormParallelConfig(
            mesh=mesh,
            activation_sharded=enable_collective_matmul,
        ),
    )
    lm_head_parallel = parallel.LinearParallelConfig(
        mesh=mesh,
        parallel_type=parallel.LinearParallelType.COLUMN,
    )

    if enable_collective_matmul:
      lm_head_parallel.collective_matmul_type = (
          parallel.CollectiveMatmulType.ALL_GATHER
      )

    self.lm_head = nn.Linear(
        config.hidden_size,
        config.vocab_size,
        parallel_config=lm_head_parallel,
    )

  def __call__(
      self,
      input_ids,
      positions,
      kv_caches,
      attn_metadata,
  ) -> tuple[jax.Array, list[nn.KVCache]]:
    hidden_states = self.embed_tokens(input_ids)
    for i in range(self.config.num_hidden_layers):
      hidden_states, kv_cache = self.layers[i](
          hidden_states,
          positions,
          kv_caches[i],
          attn_metadata,
      )
      kv_caches[i] = kv_cache

    hidden_states = self.norm(hidden_states)
    logits = self.lm_head(hidden_states)

    logits = parallel.ops.all_gather(
        logits, axis=len(logits.shape) - 1, axis_names=parallel.tp_axis_names()
    )

    return logits, kv_caches


class LlamaForCausalLM(nn.CausalLM):

  def __init__(
      self,
      config: LlamaConfig,
      parallel_config: parallel.ModelParallelConfig,
      eos: int | None = None,
      max_length: int | None = None,
  ):
    super().__init__()
    self.config = config
    self.model = LlamaModel(config=config, parallel_config=parallel_config)
    self.sampler = Sampler(eos, max_length)

  def __call__(
      self,
      input_ids: jax.Array,
      positions: jax.Array,
      kv_caches: list[nn.KVCache],
      attn_metadata: nn.AttentionMetadata,
      sampling_params: SamplingParams,
  ) -> tuple[ModelOutput, list[nn.KVCache]]:
    logits, kv_caches = self.model(
        input_ids, positions, kv_caches, attn_metadata
    )
    tokens, done = self.sampler.sample(
        logits, positions, attn_metadata, sampling_params
    )

    return postprocess(tokens, done, attn_metadata), kv_caches



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Embedding Module"""

from jax import numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P
from inference.nn import Module, Parameter
from inference.parallel import EmbeddingParallelType
from inference import parallel


class Embedding(Module):

  def __init__(
      self,
      vocab_size,
      embedding_dim,
      parallel_config: parallel.EmbeddingParallelConfig,
  ):
    super().__init__()
    self.vocab_size = vocab_size
    self.embedding_dim = embedding_dim
    self.parallel_config = parallel_config

    self.weight = Parameter(
        jnp.zeros((vocab_size, embedding_dim), dtype=jnp.bfloat16)
    )

    mesh = parallel_config.mesh
    if parallel_config.parallel_type == EmbeddingParallelType.COLUMN:
      weight_pspec = P(None, parallel.tp_axis_names())
    else:
      weight_pspec = P(None, None)

    self.weight.sharding = NamedSharding(mesh, weight_pspec)

  def __call__(self, input):
    return self.weight.value[input]


def apply_rope_embedding(input, position, theta=10000):
  emb_dim = input.shape[-1]
  fraction = jnp.arange(0, emb_dim, 2) / emb_dim
  timescale = theta**fraction
  position = position[:, None, None]
  sinusoid_inp = position / timescale
  sin = jnp.sin(sinusoid_inp).astype(input.dtype)
  cos = jnp.cos(sinusoid_inp).astype(input.dtype)
  first_half, second_half = jnp.split(input, 2, axis=-1)
  first_part = first_half * cos - second_half * sin
  second_part = second_half * cos + first_half * sin
  return jnp.concatenate((first_part, second_part), axis=-1)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""engine module"""

import enum
import dataclasses
import datetime
import math
import queue
import threading
from typing import Any
import uuid
import jax
import jax.profiler
from jax import numpy as jnp
from jax.sharding import Mesh, NamedSharding, PartitionSpec as P
import numpy as np
from inference import nn
from inference import parallel
from inference.config.config import InferenceParams
from inference.model import ModelRegistry, SamplingParams
from inference.model.llama import LlamaForCausalLM
from inference.runtime.kv_cache import *
from inference.runtime.request_type import *
from inference.runtime.kv_cache import KVCacheStorage, KVCacheManager
from inference.runtime.batch_scheduler import BatchScheduler, Schedule, SchedulePolicy
from inference.runtime.model_executor import Executor


@enum.unique
class EngineMode(enum.Enum):
  OFFLINE = enum.auto()
  ONLINE = enum.auto()


@dataclasses.dataclass
class OfflineChannel:
  req_queue: queue.Queue[OfflineRequest]
  res_queue: queue.Queue[Response]


@dataclasses.dataclass
class OnlineChannel:
  req_queue: asyncio.Queue[OnlineRequest]
  aio_loop: asyncio.AbstractEventLoop


class Engine:
  """Engine is a wrapper of the model for inference"""

  def __init__(
      self,
      mesh: Mesh,
      inference_params: InferenceParams,
      mode: EngineMode,
      channel: OfflineChannel | OnlineChannel,
      debug_mode: bool,
  ):
    print("Initializing engine")
    self.mesh = mesh
    self.inference_params = inference_params
    model_registry = ModelRegistry()

    print("-" * 40)
    print("Loading tokenizer")
    self.tokenizer = model_registry.load_tokenizer(inference_params.model_id)

    print("-" * 40)
    print("Loading model config")
    model_config = model_registry.load_model_config(inference_params.model_id)
    if debug_mode:
      model_config.num_hidden_layers = 1

    self.model: nn.CausalLM = LlamaForCausalLM(
        model_config,
        parallel.ModelParallelConfig(mesh=self.mesh),
        self.tokenizer.eos_token_id,
        self.inference_params.max_seq_length,
    )

    print("-" * 40)
    if debug_mode:
      print("Initializing random model weights to devices")
      self.weights_dict = self.model.init_weights()
    else:
      print("Loading model weights to host")
      weights_on_host = model_registry.load_weights_to_host(
          model_id=inference_params.model_id,
          num_devices=self.mesh.devices.size,
          model_config=model_config,
          dtype=jnp.bfloat16,
      )
      print("Loading model weights to devices")
      self.weights_dict = self.model.load_weights_dict(weights_on_host)

    print("-" * 40)
    print("Initializing KV cache storage/manager")
    # init kv cache
    self.kv_storage = KVCacheStorage(
        mesh=self.mesh,
        model_config=model_config,
        page_size=self.inference_params.page_size,
        hbm_utilization=self.inference_params.hbm_utilization,
    )
    self.kv_manager = KVCacheManager(
        self.kv_storage.num_hbm_pages_per_layer, self.inference_params.page_size
    )

    self.mode = mode
    self.channel = channel

    if self.mode == EngineMode.OFFLINE:
      mode = SchedulePolicy.OFFLINE
    else:
      assert self.mode == EngineMode.ONLINE
      mode = SchedulePolicy.ONLINE

    print("-" * 40)
    print("Initializing batch scheduler")
    self.scheduler = BatchScheduler(
        self.kv_manager,
        self.inference_params.batch_size,
        self.inference_params.max_seq_length,
        mode,
    )

    print("-" * 40)
    print("Initializing GenerateState")
    self.active_prefill_request = None
    self.num_pages_per_seq = math.ceil(
        self.inference_params.max_seq_length / self.inference_params.page_size
    )
    slots = queue.SimpleQueue()
    for i in range(self.inference_params.batch_size):
      slots.put(i, block=True)
    self.generate_state = GenerateState(
        token_ids=jnp.zeros(
            shape=(self.inference_params.batch_size,),
            dtype=jnp.int32,
            device=NamedSharding(self.mesh, P(None)),
        ),
        positions=jnp.full(
            shape=(self.inference_params.batch_size,),
            fill_value=-1,
            dtype=jnp.int32,
            device=NamedSharding(self.mesh, P(None)),
        ),
        page_table=jnp.full(
            shape=(
                self.inference_params.batch_size,
                self.num_pages_per_seq,
            ),
            fill_value=self.kv_manager.dummy_page_idx,
            dtype=jnp.int32,
            device=NamedSharding(self.mesh, P(None, None)),
        ),
        available_slots=slots,
        active_slot_req_map={},
    )
    self.sample_params = SamplingParams(
        temperature=jax.device_put(
            jnp.asarray((1.0), dtype=jnp.float32), NamedSharding(self.mesh, P())
        ),
        top_k=jax.device_put(
            jnp.asarray((1), dtype=jnp.int32), NamedSharding(self.mesh, P())
        ),
        rng=jax.device_put(jax.random.key(0), NamedSharding(self.mesh, P())),
    )

    self.model_executor = Executor(
        self.mesh,
        self.weights_dict,
        self.model.jittable_call,
        self.num_pages_per_seq,
        debug_mode=debug_mode,
    )

    print("-" * 40)
    print("Compiling engine")
    self.model_executor.compile(
        self.inference_params.prefill_chunk_sizes,
        self.inference_params.batch_size,
        self.inference_params.max_seq_length,
        self.inference_params.max_input_length,
        self.kv_storage.hbm_kv_caches,
        self.sample_params,
    )

    # running loop
    self.requests_dict: dict[str, Request] = {}

    print("-" * 40)
    print("Starting threads:", end="")
    if self.mode == EngineMode.OFFLINE:
      print(" dequeue,", end="")
      self._dequeue_offline_req_thread = threading.Thread(
          name="dequeue_offline_request", target=self._dequeue_offline_request
      )
    else:
      print(" dequeue,", end="")
      self._dequeue_online_req_thread = threading.Thread(
          name="_dequeue_online_request", target=self._dequeue_online_request
      )

    # TODO: Assign the max_device_requests_sem number by the
    # device spec and cost model.
    self._max_device_requests_sem = threading.Semaphore(
        self.inference_params.batch_size * 3 // 2
    )
    print(" preprocess,", end="")
    self._preprocess_queue: queue.Queue[Request] = queue.Queue()
    # TODO: Seperate the running loop with the static inference model.
    self._preprocess_thread = threading.Thread(
        name="preprocess", target=self._preprocess
    )
    # Add backpressure to prevent that the inference thread never releases
    # the GIL and keeps dispatching the device program.
    print(" postprocess,", end="")
    self._postprocess_queue: queue.Queue[PostProcessRequest] = queue.Queue(8)
    self._postprocess_thread = threading.Thread(
        name="postprocess", target=self._postprocess
    )

    print(" inference")
    self._inference_thread = threading.Thread(
        name="inference", target=self._inference
    )
    self.total_reqs = 0
    self.complete_reqs = 0
    self.start_time = None

  def start(self):
    jax.profiler.start_server(9999)

    if self.mode == EngineMode.OFFLINE:
      self._dequeue_offline_req_thread.start()
    else:
      self._dequeue_online_req_thread.start()

    self._preprocess_thread.start()
    self._postprocess_thread.start()
    self._inference_thread.start()

    self.start_time = datetime.datetime.now()
    print("-" * 40)
    print("Engine starts: ", self.start_time)

  def stop(self):
    jax.profiler.stop_server()
    # Stop listen to the queue when item is None.
    self.channel.req_queue.put(None, block=True)
    self._preprocess_queue.put(None, block=True)
    self.scheduler.enqueue_prefill_req(None)
    self.scheduler.enqueue_generate_req(None)
    self._postprocess_queue.put(None, block=True)

    if self.mode == EngineMode.OFFLINE:
      self._dequeue_offline_req_thread.join()
    else:
      self._dequeue_online_req_thread.join()

    self._preprocess_thread.join()
    self._inference_thread.join()
    self._postprocess_thread.join()

    stop_time = datetime.datetime.now()
    duration = (stop_time - self.start_time).total_seconds()
    print(f"Engine stops: {stop_time}")
    print("-" * 40)
    print(f"Engine total run time: {duration:.2f} seconds")

  def _dequeue_online_request(self):
    """ "Dequeues online requests and put them in the preprocess queue."""
    while True:
      r = self.channel.req_queue.get(block=True)
      if r is None:
        return
      assert isinstance(r, OnlineRequest)
      req = Request(
          id=uuid.uuid4().hex, prompt=r.prompt, aio_response_queue=r.res_queue
      )
      self._preprocess_queue.put(req, block=True)
      self.requests_dict[req.id] = req

  def _dequeue_offline_request(self):
    """ "Dequeues offline requests and put them in the preprocess queue."""
    while True:
      r = self.channel.req_queue.get(block=True)
      if r is None:
        return
      assert isinstance(r, OfflineRequest)
      req = Request(id=uuid.uuid4().hex, prompt=r.prompt)
      self._preprocess_queue.put(req, block=True)
      self.requests_dict[req.id] = req

  def _encode(self, prompt: str) -> tuple[np.ndarray, list[int]]:
    """Encodes the text prompt and trims/pads it to the max input length."""
    token_id_list = self.tokenizer.encode(prompt)
    token_id_array = np.asarray(token_id_list)
    length = token_id_array.size
    assert length == len(token_id_list)
    bound = self.inference_params.max_input_length
    if length == bound:
      return token_id_array, token_id_list
    elif length < bound:
      return np.pad(token_id_array, (0, bound - length)), token_id_list
    else:
      return token_id_array[-bound:], token_id_list[-bound:]

  def _select_chunk_size(self, token_len: int) -> int:
    """Selects a prefill chunk size that is big enough."""
    for size in self.inference_params.prefill_chunk_sizes:  # ascending
      if token_len <= size:
        return size
    return self.inference_params.prefill_chunk_sizes[-1]

  def _build_prefill_request(self, req: Request) -> PrefillRequest:
    """Builds a token prefill request from the original text prompt."""
    padded_token_ids, req.prompt_token_ids = self._encode(req.prompt)
    chunk_size = self._select_chunk_size(len(req.prompt_token_ids))
    pi = self.kv_manager.dummy_page_idx
    dummy_page_indices = [pi] * self.num_pages_per_seq
    device_tokens = jax.device_put(
        padded_token_ids, NamedSharding(self.mesh, P(None))  # no sharding
    )
    positions = np.arange(0, device_tokens.shape[0])
    device_positions = jax.device_put(
        positions, NamedSharding(self.mesh, P(None))  # no sharding
    )
    return PrefillRequest(
        id=req.id,
        unpadded_token_ids=req.prompt_token_ids,
        chunk_idx=0,
        chunk_size=chunk_size,
        page_indices=dummy_page_indices,
        device_token_ids=device_tokens,
        device_positions=device_positions,
    )

  def _preprocess(self) -> jax.Array:
    """Converts prompts to prefill requests and give them to the scheduler."""
    while True:
      req = self._preprocess_queue.get(block=True)
      if req is None:
        return
      assert isinstance(req, Request)
      # Don't put too many pending requests to the HBM.
      self._max_device_requests_sem.acquire()
      prefill = self._build_prefill_request(req)
      self.scheduler.enqueue_prefill_req(prefill)

  def _inference(self) -> jax.Array:
    while True:
      schedule = self.scheduler.schedule(
          self.active_prefill_request, self.generate_state
      )
      if schedule is None:
        return

      assert isinstance(schedule, Schedule)
      input = self.model_executor.prepare_input_and_update_generate_state(
          schedule,
          self.generate_state,
          self.kv_storage.hbm_kv_caches,
          self.sample_params,
          self.inference_params.batch_size,
      )
      output, self.kv_storage.hbm_kv_caches = self.model_executor.execute(input)

      # Prepare for next iteration and post-processed request.
      post_req = PostProcessRequest(
          prefill_request_id=None,
          prefill_token_id=output.prefill_token,
          prefill_done=output.prefill_done,
          generate_active_slots=[],
          generate_active_request_ids=[],
          generate_token_ids=output.generate_tokens,
          generate_done=output.generate_done,
      )

      if schedule.schedule_prefill:
        prefill = schedule.prefill_request
        prefill.chunk_idx += 1
        start_idx = prefill.chunk_idx * prefill.chunk_size
        prefill_length = len(prefill.unpadded_token_ids)
        if start_idx < prefill_length:
          self.active_prefill_request = prefill
        else:
          self.active_prefill_request = None
          post_req.prefill_request_id = schedule.prefill_request.id
          generate_req = GenerateRequest(
              id=prefill.id,
              slot=-1,
              pos=prefill_length,
              page_indices=prefill.page_indices,
              device_prefill_token_id=output.prefill_token,
          )
          self.scheduler.enqueue_generate_req(generate_req)

      if schedule.schedule_generate:
        self.generate_state.token_ids = output.generate_tokens
        self.generate_state.positions = output.generate_next_pos
        with self.generate_state.map_mutex:
          for (
              slot,
              processed_gr,
          ) in self.generate_state.active_slot_req_map.items():
            processed_gr.pos += 1
            post_req.generate_active_slots.append(slot)
            post_req.generate_active_request_ids.append(processed_gr.id)

      self._postprocess_queue.put(post_req, block=True)

  def _postprocess(self) -> str:
    while True:
      post = self._postprocess_queue.get(block=True)
      if post is None:
        return

      assert isinstance(post, PostProcessRequest)
      post.prefill_token_id = np.asarray(post.prefill_token_id).item()
      post.prefill_done = np.asarray(post.prefill_done).item()
      post.generate_token_ids = np.asarray(post.generate_token_ids).tolist()
      post.generate_done = np.asarray(post.generate_done).tolist()

      # Free finished slot.
      if len(post.generate_active_request_ids) > 0:
        with self.generate_state.map_mutex:
          slot_to_del = []
          for slot in self.generate_state.active_slot_req_map.keys():
            if post.generate_done[slot]:
              self.generate_state.available_slots.put(slot, block=True)
              pages_to_free = self.generate_state.active_slot_req_map[
                  slot
              ].page_indices
              self.kv_manager.free_hbm_pages(pages_to_free)
              slot_to_del.append(slot)
          for slot in slot_to_del:
            del self.generate_state.active_slot_req_map[slot]

      # Return generated tokens to the client.
      if post.prefill_request_id:
        req = self.requests_dict[post.prefill_request_id]
        req.generated_token_ids.append(post.prefill_token_id)
        generated_text = self.tokenizer._convert_id_to_token(
            post.prefill_token_id
        ).replace("▁", " ")
        req.generated_text += generated_text

        if self.mode == EngineMode.ONLINE:
          self.channel.aio_loop.call_soon_threadsafe(
              req.aio_response_queue.put_nowait,
              Response(generated_text, post.prefill_token_id),
          )

        if post.prefill_done:
          req.completed = True
          if self.mode == EngineMode.ONLINE:
            self.channel.aio_loop.call_soon_threadsafe(
                req.aio_response_queue.put_nowait,
                Response(generated_text, post.prefill_token_id),
            )

          else:
            self.channel.res_queue.put_nowait(
                Response(
                    generated_text=req.generated_text,
                    generated_tokens=req.generated_token_ids,
                    input_tokens=req.prompt_token_ids,
                )
            )

          del self.requests_dict[post.prefill_request_id]
          self._max_device_requests_sem.release()

      for slot, req_id in zip(
          post.generate_active_slots, post.generate_active_request_ids
      ):
        if req_id not in self.requests_dict:
          continue
        req = self.requests_dict[req_id]

        req.generated_token_ids.append(post.generate_token_ids[slot])
        generated_text = self.tokenizer._convert_id_to_token(
            post.generate_token_ids[slot]
        ).replace("▁", " ")
        req.generated_text += generated_text

        if self.mode == EngineMode.ONLINE:
          self.channel.aio_loop.call_soon_threadsafe(
              req.aio_response_queue.put_nowait,
              Response(
                  generated_text=generated_text,
                  generated_tokens=post.generate_token_ids[slot],
              ),
          )

        if post.generate_done[slot]:
          req.completed = True
          if self.mode == EngineMode.ONLINE:
            self.channel.aio_loop.call_soon_threadsafe(
                req.aio_response_queue.put_nowait,
                None,
            )
          else:
            self.channel.res_queue.put_nowait(
                Response(
                    generated_text=req.generated_text,
                    generated_tokens=req.generated_token_ids,
                    input_tokens=req.prompt_token_ids,
                )
            )

          self._max_device_requests_sem.release()
          del self.requests_dict[req_id]



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Hugging Face Llama2 ckpt conversion utility function."""

from jax import numpy as jnp
from transformers import LlamaConfig
from .util import convert_to_jax_array_on_cpu


def merge_gate_up_proj_weights(gate, up, num_devices):
  col_chunk_size = gate.shape[1] // num_devices
  weight = jnp.concatenate(
      [gate[:, :col_chunk_size], up[:, :col_chunk_size]], axis=-1
  )
  for i in range(1, num_devices):
    index = col_chunk_size * i
    weight = jnp.concatenate(
        [
            weight,
            gate[:, index : index + col_chunk_size],
            up[:, index : index + col_chunk_size],
        ],
        axis=-1,
    )

  return weight


def convert_hf_llama(torch_weight_state, num_devices, config: LlamaConfig):
  jax_weight_state = {
      "embed_tokens": {
          "weight": convert_to_jax_array_on_cpu(
              torch_weight_state["model.embed_tokens.weight"]
          ),
      },
      "layers": {},
      "norm": {
          "weight": convert_to_jax_array_on_cpu(
              torch_weight_state["model.norm.weight"]
          ),
      },
      "lm_head": {
          "weight": convert_to_jax_array_on_cpu(
              torch_weight_state["lm_head.weight"].T
          ),
      },
  }
  del torch_weight_state["model.embed_tokens.weight"]
  del torch_weight_state["model.norm.weight"]
  del torch_weight_state["lm_head.weight"]
  for i in range(config.num_hidden_layers):
    gate_up_proj_weight = merge_gate_up_proj_weights(
        gate=convert_to_jax_array_on_cpu(
            torch_weight_state[f"model.layers.{i}.mlp.gate_proj.weight"].T
        ),
        up=convert_to_jax_array_on_cpu(
            torch_weight_state[f"model.layers.{i}.mlp.up_proj.weight"].T
        ),
        num_devices=num_devices,
    )
    del torch_weight_state[f"model.layers.{i}.mlp.gate_proj.weight"]
    del torch_weight_state[f"model.layers.{i}.mlp.up_proj.weight"]
    jax_weight_state["layers"][i] = {
        "self_attn": {
            "q_proj": {
                "weight": convert_to_jax_array_on_cpu(
                    torch_weight_state[
                        f"model.layers.{i}.self_attn.q_proj.weight"
                    ].T
                ),
            },
            "k_proj": {
                "weight": convert_to_jax_array_on_cpu(
                    torch_weight_state[
                        f"model.layers.{i}.self_attn.k_proj.weight"
                    ].T
                ),
            },
            "v_proj": {
                "weight": convert_to_jax_array_on_cpu(
                    torch_weight_state[
                        f"model.layers.{i}.self_attn.v_proj.weight"
                    ].T
                ),
            },
            "o_proj": {
                "weight": convert_to_jax_array_on_cpu(
                    torch_weight_state[
                        f"model.layers.{i}.self_attn.o_proj.weight"
                    ].T
                ),
            },
        },
        "ffw": {
            "gate_up_proj": {
                "weight": gate_up_proj_weight,
            },
            "down_proj": {
                "weight": convert_to_jax_array_on_cpu(
                    torch_weight_state[
                        f"model.layers.{i}.mlp.down_proj.weight"
                    ].T
                ),
            },
        },
        "input_layernorm": {
            "weight": convert_to_jax_array_on_cpu(
                torch_weight_state[f"model.layers.{i}.input_layernorm.weight"]
            ),
        },
        "post_attention_layernorm": {
            "weight": convert_to_jax_array_on_cpu(
                torch_weight_state[
                    f"model.layers.{i}.post_attention_layernorm.weight"
                ]
            ),
        },
    }
    del torch_weight_state[f"model.layers.{i}.self_attn.q_proj.weight"]
    del torch_weight_state[f"model.layers.{i}.self_attn.k_proj.weight"]
    del torch_weight_state[f"model.layers.{i}.self_attn.v_proj.weight"]
    del torch_weight_state[f"model.layers.{i}.self_attn.o_proj.weight"]
    del torch_weight_state[f"model.layers.{i}.mlp.down_proj.weight"]
    del torch_weight_state[f"model.layers.{i}.input_layernorm.weight"]
    del torch_weight_state[f"model.layers.{i}.post_attention_layernorm.weight"]

  del torch_weight_state

  jax_causal_lm_weight_state = {}
  jax_causal_lm_weight_state["model"] = jax_weight_state
  return jax_causal_lm_weight_state



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JetStream Http API server."""

import json
import logging
import time
from typing import Sequence
from absl import app as abslapp
from absl import flags
from fastapi import APIRouter, Response
import fastapi
from fastapi.responses import StreamingResponse
from prometheus_client import start_http_server
import uvicorn
from google.protobuf.json_format import Parse

from jetstream.core import config_lib, orchestrator, server_lib
from jetstream.core.metrics.prometheus import JetstreamMetricsCollector
from jetstream.core.proto import jetstream_pb2
from jetstream.entrypoints.config import get_server_config
from jetstream.entrypoints.http.protocol import DecodeRequest
from jetstream.entrypoints.http.utils import proto_to_json_generator

flags.DEFINE_string("host", "0.0.0.0", "server host address")
flags.DEFINE_integer("port", 8080, "http server port")
flags.DEFINE_string(
    "config",
    "InterleavedCPUTestServer",
    "available servers",
)
flags.DEFINE_integer(
    "prometheus_port",
    9988,
    "prometheus_port",
)

llm_orchestrator: orchestrator.LLMOrchestrator

# Define Fast API endpoints (use llm_orchestrator to handle).
router = APIRouter()


@router.get("/")
def root():
  """Root path for Jetstream HTTP Server."""
  return Response(
      content=json.dumps({"message": "JetStream HTTP Server"}, indent=4),
      media_type="application/json",
  )


@router.post("/v1/generate")
async def generate(request: DecodeRequest):
  start_time = time.perf_counter()
  proto_request = Parse(request.json(), jetstream_pb2.DecodeRequest())
  metadata = jetstream_pb2.DecodeRequest.Metadata()
  metadata.start_time = start_time
  proto_request.metadata.CopyFrom(metadata)
  generator = llm_orchestrator.Decode(proto_request)
  return StreamingResponse(
      content=proto_to_json_generator(generator), media_type="text/event-stream"
  )


@router.get("/v1/health")
async def health() -> Response:
  """Health check."""
  response = await llm_orchestrator.HealthCheck(
      jetstream_pb2.HealthCheckRequest()
  )
  return Response(
      content=json.dumps({"is_live": str(response.is_live)}, indent=4),
      media_type="application/json",
      status_code=200,
  )


def server(argv: Sequence[str]):
  # Init Fast API.
  app = fastapi.FastAPI()
  app.include_router(router)

  # Init LLMOrchestrator which would be the main handler in the api endpoints.
  devices = server_lib.get_devices()
  print(f"devices: {devices}")
  server_config = get_server_config(flags.FLAGS.config)
  print(f"server_config: {server_config}")
  del argv

  metrics_server_config: config_lib.MetricsServerConfig | None = None
  # Setup Prometheus server
  metrics_collector: JetstreamMetricsCollector = None
  if flags.FLAGS.prometheus_port != 0:
    metrics_server_config = config_lib.MetricsServerConfig(
        port=flags.FLAGS.prometheus_port
    )
    logging.info(
        "Starting Prometheus server on port %d", metrics_server_config.port
    )
    start_http_server(metrics_server_config.port)
    metrics_collector = JetstreamMetricsCollector(
        model_name=metrics_server_config.model_name
    )
  else:
    logging.info(
        "Not starting Prometheus server: --prometheus_port flag not set"
    )

  global llm_orchestrator
  llm_orchestrator = orchestrator.LLMOrchestrator(
      driver=server_lib.create_driver(
          config=server_config,
          devices=devices,
          metrics_collector=metrics_collector,
      )
  )

  # Start uvicorn http server.
  uvicorn.run(
      app, host=flags.FLAGS.host, port=flags.FLAGS.port, log_level="info"
  )


if __name__ == "__main__":
  # Run Abseil app w flags parser.
  abslapp.run(server)



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os

# Allows faster HF download
os.environ["HF_HUB_ENABLE_HF_TRANSFER"] = "1"
os.environ["KERAS_BACKEND"] = "jax"
os.environ["JAX_PLATFORMS"] = ""


from pathlib import Path
import subprocess
import importlib.metadata
import sys
import warnings


def check_python_version():
    """
    Check if Python version is 3.11 or above, and issue a warning if it's not.
    """
    python_version = sys.version_info

    if python_version.major < 3 or (
        python_version.major == 3 and python_version.minor < 11
    ):
        warning_message = (
            f"Warning: Your current Python version is {python_version.major}.{python_version.minor}. "
            "Kithara requires Python 3.11 or above."
        )
        warnings.warn(warning_message, RuntimeWarning)
        return False
    return True


check_python_version()


def _install_maxtext():
    try:
        importlib.metadata.version("maxtext")
    except importlib.metadata.PackageNotFoundError:
        try:
            print(
                "Installing MaxText... This should only happen once when Kithara is first initiated."
            )
            maxtext_path = Path(
                os.path.join(os.path.dirname(Path(__file__)), "model/maxtext/maxtext")
            )
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "-e",
                    str(maxtext_path),
                    "--no-deps",
                ]
            )
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "pathwaysutils@git+https://github.com/google/pathways-utils.git",
                ]
            )
            print("MaxText installed successfully")
        except Exception as e:
            print(f"Failed to install maxtext: {e}")

    maxtext_dir = Path(__file__).parent / "model/maxtext/maxtext/MaxText"
    sys.path.append(str(maxtext_dir))


def _install_jetstream():
    try:
        importlib.metadata.version("google-jetstream")
    except importlib.metadata.PackageNotFoundError:
        try:
            print(
                "Installing JetStream... This should only happen once when Kithara is first initiated."
            )
            jetstream_path = Path(
                os.path.join(os.path.dirname(Path(__file__)), "model/maxtext/JetStream")
            )
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "-e",
                    str(jetstream_path),
                    "--no-deps",
                ]
            )
            print("JetStream installed successfully")
        except Exception as e:
            print(f"Failed to install JetStream: {e}")

    jetstream_dir = Path(__file__).parent / "model/maxtext/JetStream"
    sys.path.append(str(jetstream_dir))


def _install_seqio():
    """We need to install seqio separate from the rest of packages in
    project.toml for now due to endless backtracking during
    pip dependency resolution. We are installing the seqio package upon
    kithara initialization with the --no-deps flag.
    """
    try:
        importlib.metadata.version("seqio")
    except importlib.metadata.PackageNotFoundError:
        try:
            print(
                "Installing Seqio... This should only happen once when Kithara is first initiated."
            )
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "seqio==0.0.19",
                    "--no-deps",
                ]
            )
            print("Seqio installed successfully")
        except Exception as e:
            print(f"Failed to install Seqio: {e}")


_install_maxtext()
_install_jetstream()
_install_seqio()

from kithara.dataset import Dataloader, SFTDataset, TextCompletionDataset
from kithara.trainer import Trainer
from kithara.callbacks import Checkpointer, Profiler
import jax
from kithara.utils.gcs_utils import find_cache_root_dir
from kithara.model import KerasHubModel, MaxTextModel, Model
from kithara.distributed import ShardingStrategy, PredefinedShardingStrategy

# Cache JAX compilation to speed up future runs. You should notice
# speedup of training step up on the second run of this script.
jax_cache_dir = os.path.join(find_cache_root_dir(), "jax_cache")
jax.config.update("jax_compilation_cache_dir", jax_cache_dir)

print(f"JAX compilation cached at {jax_cache_dir}")

from kithara.utils.logging_utils import print_kithara_logo_and_platform_info

try:
    print_kithara_logo_and_platform_info()
except Exception as e:
    pass
